/*
 Copyright 2012, S.S. Papadopulos & Assoc. Inc.

    This file is part of GENIE.

    The GENIE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    GENIE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GENIE.  If not, see<http://www.gnu.org/licenses/>.
*/

// cmuffels, Genie version 1, 2012

#include <iostream>

#include "defin.h"
#include "connector.h"
#include "header.h"

using namespace std;

//****************************************************************************
void CONNECTOR::set_host(HOST *_host)
//****************************************************************************
//CTM Mar 2011

/*
      Set host through which connections will be managed
*/

{
  host=_host;
}

//****************************************************************************
void CONNECTOR::set_queue_slaves(list<CLIENT*> *_slaves)
//****************************************************************************
//CTM Mar 2011

/*
      Set list used to track new slave connections
*/

{
  slaves=_slaves;
}

//****************************************************************************
int CONNECTOR::execute()
//****************************************************************************
//CTM Mar 2011

/*
      Execute thread process
*/

{

  DWORD tr_Wait,m_Wait;
  bool endthread=false;

  try
  {

    // duplicate handles of primary thread
    if(!DuplicateHandle(primary,ph_thComm,GetCurrentProcess(),&h_thComm,
                        0,true,DUPLICATE_SAME_ACCESS)) throw -1;
    if(!DuplicateHandle(primary,pm_thComm,GetCurrentProcess(),&m_thComm,
                        0,true,DUPLICATE_SAME_ACCESS)) throw -2;
    if(!DuplicateHandle(primary,ph_thEnd,GetCurrentProcess(),&h_thEnd,
                        0,true,DUPLICATE_SAME_ACCESS)) throw -3;

    while(1)
    {

      // check whether to terminate thread
      if(endthread)
      {
        CloseHandle(h_thEnd);
        break;
      }

      // check whether terminate command received
      tr_Wait=WaitForSingleObject(h_thEnd,COMM_WAIT_TIME*CLOCKS_PER_SEC/100);
      if(tr_Wait==WAIT_OBJECT_0)
      {
        CloseHandle(h_thEnd);
        break;
      }

      switch(host->waitforcommunication(COMM_WAIT_TIME,0))
      {
      case SOCKET_ERROR:
        // error waiting on host socket
        throw ERROR_SOCKET_WAIT;
        break;
      case SELECT_TIMEOUT:
        break;
      default:
        //cout<<"new connection."<<'\n';
        slave=new CLIENT;
        slave->setsocket(host->acceptnewconnection());
        slave->name="NOT_SET";
        slave->type=TYPE_NOT_SET;
        slave->status=WAITING_FOR_RUN;
        slave->init_timers();
        slave->terminate=false;
        if(slave->sock==SOCKET_ERROR) throw ERROR_SOCKET_ACCEPT;
        m_Wait=WaitForSingleObject(m_thComm,INFINITE);
        if(m_Wait!=WAIT_OBJECT_0) throw ERROR_EVENT_WAIT;
        //cout<<"have slaves queue."<<'\n';
        slaves->push_back(slave);
        if(!ReleaseMutex(m_thComm)) throw ERROR_EVENT_RELEASE;
        break;
      } // end switch wait for new connection

    }
    return 1;
  }

  catch(int err)
  {
    switch(err)
    {
    case -1:
      cout<<"\n\n   Error duplicating misc. communication flag."<<'\n';
      break;
    case -2:
      cout<<"\n\n   Error duplicating misc. shared access flag."<<'\n';
      break;
    case -3:
      cout<<"\n\n   Error duplicating termination flag."<<'\n';
      break;
    }
    cout<<" *** THREAD :"<<threadid<<'\n';
    return 0;
  }

}
