/*
 Copyright 2012, S.S. Papadopulos & Assoc. Inc.

    This file is part of GENIE.

    The GENIE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    GENIE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GENIE.  If not, see<http://www.gnu.org/licenses/>.
*/

// cmuffels, Genie version 1, 2012

#include <iostream>
#include <list>
#include <map>
#include <set>

#include "defin.h"
#include "host.h"
#include "client.h"
#include "receivor.h"
#include "connector.h"
#include "sendor.h"
#include "terminator.h"
#include "modelrun.h"
#include "modelresult.h"
#include "runmanager.h"
using namespace std;

////----------------------------------------------------------------------------
//struct cmp_socket
//{
//  bool operator() (CLIENT *a,CLIENT *b)
//  {
//      return a->sock<b->sock;
//  }
//};

////----------------------------------------------------------------------------
//struct cmp_client_runtime
//{
//  bool operator() (RECEIVOR *a,RECEIVOR *b)
//  {
//      return a->slave->median_runtime()<b->slave->median_runtime();
//  }
//};
//
////----------------------------------------------------------------------------
//struct cmp_client_timeleft
//{
//  bool operator() (RECEIVOR *a,RECEIVOR *b)
//  {
//      return !(a->slave->expected_timeleft()<b->slave->expected_timeleft());
//  }
//};

//****************************************************************************
int main(int argc,     //number of args in argv
         char *argv[]) // Array of arguments
//****************************************************************************
/*

     C.Muffels, S.S. Papadopulos & Assoc. Inc.
     W.Shreuder, Principia Mathematica
     Inspired by a run manager by J.Doherty and BeoPEST by W.Shreuder

     Run Manager: Each slave is managed by its own thread.
                  A separate thread is used to handle all client connection
                     request
                  The primary thread waits on a series of signals from these
                     client threads and processes them accordingly:
                     1) New slave connection
                     2) Model connected
                     3) Run received


     Version 1.0.0.0  Dec. 2010
     Version 1.0.0.1  Apr. 2011  -- Load balancing functionality
     Version 1.0.0.2  Apr. 2011  -- Connection loss handling

*/

{

  //--- MISC
  bool endprogram=false;
  bool modelconnected=false;
  DWORD m_wait;
  //DWORD th_ExitCode;
  RUN_MANAGER rm;

  HOST host;
  CLIENT *modelclient;

  CONNECTOR connector;
  SENDOR sendor;
  TERMINATOR terminator;

  HANDLE h_thComm;
  HANDLE h_thSend;
  HANDLE h_thTerm;
  HANDLE m_thTerm;
  HANDLE m_thRuns;
  HANDLE m_thSend;
  HANDLE m_thSlaves;

  RECEIVOR *newrecv;
  RECEIVOR *modelthread;

  list<CLIENT*> *clients_new;
  list<CLIENT*> *clients_connected;
  list<CLIENT*>::iterator client;

  list<RECEIVOR*> *threads;
  list<RECEIVOR*> *threads_end;
  list<RECEIVOR*>::iterator thread;

  WSADATA wsaData;

  list<MODEL_RUN*> *runs_new;
  list<MODEL_RUN*> *runs_complete;
  //list<MODEL_RUN*> *runs_tosend;
  list<MODEL_RUN*>::iterator run;
  list<MODEL_RESULT*> *results_new;
  list<MODEL_RESULT*> *results_tosend;
  list<MODEL_RESULT*>::iterator result;

  runs_new=new list<MODEL_RUN*>;
  runs_complete=new list<MODEL_RUN*>;
  results_tosend=new list<MODEL_RESULT*>;
  results_new=new list<MODEL_RESULT*>;

  map<CLIENT*,bool> slaveconnected;
  map<CLIENT*,bool>::iterator isc;

  threads=new list<RECEIVOR*>;
  threads_end=new list<RECEIVOR*>;

  clients_new=new list<CLIENT*>;
  clients_connected=new list<CLIENT*>;

  SECURITY_DESCRIPTOR sd={0};
  SECURITY_ATTRIBUTES saAttr;
  InitializeSecurityDescriptor(&sd,SECURITY_DESCRIPTOR_REVISION);
  SetSecurityDescriptorDacl(&sd,true,NULL,false);
  saAttr.nLength=sizeof(SECURITY_ATTRIBUTES);
  saAttr.lpSecurityDescriptor=&sd;
  saAttr.bInheritHandle=true;

  h_thComm=CreateEvent(&saAttr,false,false,"ThreadComm");
  if(h_thComm==NULL) throw ERROR_EVENT_INIT;
  m_thRuns=CreateMutex(&saAttr,false,"RunsQueue");
  if(m_thRuns==NULL) throw ERROR_EVENT_INIT;

  h_thSend=CreateEvent(&saAttr,false,false,"SendRuns");
  if(h_thSend==NULL) throw ERROR_EVENT_INIT;
  h_thTerm=CreateEvent(&saAttr,false,false,"TerminateThreads");
  if(h_thTerm==NULL) throw ERROR_EVENT_INIT;

  m_thSlaves=CreateMutex(&saAttr,false,"SlavesQueue");
  if(m_thSlaves==NULL) throw ERROR_EVENT_INIT;
  m_thSend=CreateMutex(&saAttr,false,"SendQueue");
  if(m_thSend==NULL) throw ERROR_EVENT_INIT;
  m_thTerm=CreateMutex(&saAttr,false,"TerminateQueue");
  if(m_thTerm==NULL) throw ERROR_EVENT_INIT;

  modelclient=NULL;
  modelthread=NULL;

  try
  {

    // initialize winsock
    if(WSAStartup(WINSOCK_VERSION,&wsaData)!=0) throw ERROR_WSASTARTUP;

    // initialize and start host
    if(!host.initialize(argc,argv)) throw ERROR_NODE_INIT;
    host.init_timers();
    if(!host.start()) throw ERROR_NODE_START;

    cout<<"\n   Initializing THREADS --->"<<'\n';

    // initialize and start thread to handle incoming connection requests
    cout<<"   - CONNECTOR (Receives new connections)... ";
    if(!connector.initialize()) throw ERROR_THREAD_START;
    connector.set_host(&host);
    connector.set_event_comm(h_thComm);
    connector.set_mutex_comm(m_thSlaves);
    connector.set_queue_slaves(clients_new);
    connector.start();
    cout<<"Complete."<<'\n';

    // initialize and start thread to return results to calling program
    cout<<"   - SENDOR (Returns results to calling program)... ";
    if(!sendor.initialize()) throw ERROR_THREAD_START;
    sendor.set_event_comm(h_thSend);
    sendor.set_mutex_comm(m_thSend);
    sendor.set_queue_results(results_tosend);
    sendor.start();
    cout<<"Complete."<<'\n';

    // initialize and start thread used to terminate threads
    cout<<"   - TERMINATOR (Terminates disconnected slave-computers)... ";
    if(!terminator.initialize()) throw ERROR_NODE_START;
    terminator.set_event_comm(h_thTerm);
    terminator.set_mutex_comm(m_thTerm);
    terminator.start();
    cout<<"Complete."<<'\n';

    // initialize run manager
    cout<<"\n   Initializing RUN MANAGER."<<'\n';
    rm.initialize();

    // wait for something to happen!
    while(1)
    {

      // check if calling program has signalled termination of GMAN
      if(modelconnected)
      {
        if(modelclient->terminate)
          endprogram=true;
      }

      // check if GMAN is to be terminated
      if(endprogram) break;

      // check if there are new slaves
      if(WaitForSingleObject(m_thSlaves,0)==WAIT_OBJECT_0)
      {
        //cout<<"check if there are new slaves."<<'\n';
        for(client=clients_new->begin();client!=clients_new->end();++client)
        {
          // create thread to handle receipt of messages from this slave
          newrecv=new RECEIVOR;
          if(!newrecv->initialize()) throw ERROR_THREAD_START;
          newrecv->set_slave(*client);
          newrecv->set_event_comm(h_thComm);
          newrecv->set_mutex_comm(m_thRuns);
          newrecv->set_queue_runs(runs_new);
          threads->push_back(newrecv);
          newrecv->start();
          clients_connected->push_back(*client);
          //cout<<"Connected a new slave: "<<newrecv->threadid<<'\n';
        }
        clients_new->clear();
        if(!ReleaseMutex(m_thSlaves)) throw ERROR_EVENT_RELEASE;
      }
      // check whether type of slave is known
      if(clients_connected->size()>0)
      {
        client=clients_connected->begin();
        while(client!=clients_connected->end())
        {
          // check type of Slave
          if((*client)->type==TYPE_SLAVE)
          {
            cout<<"\n   Connected new CLIENT: "<<(*client)->name<<'\n';
            rm.add_new_client(*client);
            client=clients_connected->erase(client);
          }
          else if((*client)->type==TYPE_MODEL)
          {
            cout<<"\n   Connected new CLIENT: "<<(*client)->name<<'\n';
            modelclient=*client;
            modelconnected=true;
            client=clients_connected->erase(client);
            // find which thread controls this client
            thread=threads->begin();
            while(thread!=threads->end())
            {
              if((*thread)->slave->sock==modelclient->sock)
              {
                modelthread=*thread;
                threads->erase(thread);
                break;
              }
              else
                ++thread;
            }
          }
          else
            ++client;
        }
      }

      //// check if there is a request for the number of connected slaves
      //if(modelconnected)
      //{
      //  if(modelclient->status==RETURN_NSLAVE)
      //  {
      //    modelclient->sendnslave(clients_connected->size());
      //  }
      //}

      // check if there are new runs
      if(runs_new->size()>0)
      {
        if(WaitForSingleObject(m_thRuns,0)==WAIT_OBJECT_0)
        {
          for(run=runs_new->begin();run!=runs_new->end();++run)
          {
            cout<<"\n   Have new RUN: "<<*(*run)->id<<'\n';
            rm.add_new_run(*run);
          }
          runs_new->clear();
          if(!ReleaseMutex(m_thRuns)) throw ERROR_EVENT_RELEASE;
        }
      }

      // check if model has disconnected
      if(modelconnected)
      {
        if(modelclient->status==CLIENT_LOST)
        {
          modelconnected=false;
          modelclient=NULL;
          threads_end->push_back(modelthread);
          modelthread=NULL;
          cout<<"\n   The Calling Program has DISCONNECTED"<<'\n';
        }
      }
      
      // start any waiting runs on available clients
      rm.start_waiting_runs();

      // handle any disconnected clients
      rm.check_client_status(slaveconnected);
      if(slaveconnected.size()>0)
      {
        for(thread=threads->begin();thread!=threads->end();++thread)
        {
          isc=slaveconnected.find((*thread)->slave);
          if(isc!=slaveconnected.end())
            if(!slaveconnected[(*thread)->slave])
            {
              cout<<"\n   Disconnected CLIENT: "<<(*thread)->slave->name<<'\n';
              threads_end->push_back(*thread);
            }
        }
      }

      // check if any runs are complete
      for(thread=threads->begin();thread!=threads->end();++thread)
      {
        if(*(*thread)->haveresult)
        {
          results_new->push_back((*thread)->get_result());
          *(*thread)->haveresult=false;
        }
      }
      rm.check_run_status();

      // send completed runs back to calling program
      if(modelconnected&&results_new->size()>0)
      {
        sendor.set_calling_program(modelclient);
        if(!ResetEvent(h_thSend)) throw ERROR_EVENT_RELEASE;
        if(WaitForSingleObject(m_thSend,0)==WAIT_OBJECT_0)
        {
          for(result=results_new->begin();result!=results_new->end();++result)
          {
            cout<<"\n   Completed RUN: "<<*(*result)->id<<'\n';
            results_tosend->push_back(*result);
            }
          if(!SetEvent(h_thSend)) throw ERROR_EVENT_SET;
          results_new->clear();
          if(!ReleaseMutex(m_thSend)) throw ERROR_EVENT_RELEASE;
        }
      }

      // Terminate any disconnected threads
      if(threads_end->size()!=0)
      {
        if(WaitForSingleObject(m_thTerm,0)==WAIT_OBJECT_0)
        {
          terminator.set_threads_end(*threads_end);
          ReleaseMutex(m_thTerm);
          if(!SetEvent(h_thTerm)) throw ERROR_EVENT_SET;
          threads_end->clear();
        }
      }

      //// check if a run can be finished faster on an idle machine
      ////cout<<"check if run can be finished on a faster machine."<<'\n';
      //if(runs_todo->size()==0&&threads_waiting->size()>0)
      //{
      //  // for each run currently being executed get the expected time left
      //  dur=new multiset<RECEIVOR*,cmp_client_timeleft>;
      //  for(ithWork=threads_working->begin();ithWork!=threads_working->end();++ithWork)
      //    dur->insert(*ithWork);
      //  // get list of slaves with at least 1 completed run
      //  run_completed=new list<RECEIVOR*>;
      //  for(ithWait=threads_waiting->begin();ithWait!=threads_waiting->end();++ithWait)
      //    if((*ithWait)->slave->run_end->size()!=0)
      //      run_completed->push_back(*ithWait);
      //  // if none of the waiting threads have completed a run then start
      //  // a run on the slave whose rating is at least 50% more than that
      //  // of the slave currently executing the run
      //  if(run_completed->size()==0)
      //  {
      //    for(idur=dur->rbegin();idur!=dur->rend();++idur)
      //    {
      //      ithWait=threads_waiting->begin();
      //      while(ithWait!=threads_waiting->end())
      //      {
      //        if((*ithWait)->slave->rating>1.5*(*idur)->slave->rating)
      //        {
      //          // start run on this machine
      //          (*ithWait)->set_run((*idur)->get_run());
      //          mRuns[(*idur)->get_run()].push_back(*ithWait);
      //          (*ithWait)->slave->run_start->push_back(clock());
      //          //(*thread)->run->send((*thread)->slave->sock());
      //          (*idur)->get_run()->send((*ithWait)->slave->sock);
      //          threads_working->insert(*ithWait);
      //          threads_waiting->erase(ithWait);
      //          break;
      //        }
      //        ++ithWait;
      //      }
      //    }
      //  }
      //  // otherwise start this run on a machine that can be expected to complete it faster
      //  else
      //  {
      //    for(idur=dur->rbegin();idur!=dur->rend();++idur)
      //    {
      //      irc=run_completed->begin();
      //      while(irc!=run_completed->end())
      //      {
      //        if((*irc)->slave->median_runtime()<(*idur)->slave->expected_timeleft())
      //        {
      //          // start run on this machine
      //          (*irc)->set_run((*idur)->get_run());
      //          mRuns[(*idur)->get_run()].push_back(*irc);
      //          (*irc)->slave->run_start->push_back(clock());
      //          (*idur)->get_run()->send((*irc)->slave->sock);
      //          threads_working->insert(*irc);
      //          threads_waiting->erase(*irc);
      //          break;
      //        }
      //        ++irc;
      //      }
      //    }
      //  }
      //  // delete containers
      //  run_completed->clear();
      //  delete run_completed;
      //  run_completed=NULL;
      //  dur->clear();
      //  delete dur;
      //  dur=NULL;
      //}

    }
        
    throw NORMAL_TERMINATION;
  }

  catch (int err)
  {

    cout<<"\n\n *** Terminating Program... **********\n"<<'\n';

    // Kill the CONNECTOR thread
    cout<<" *** Terminating CONNECTOR thread."<<'\n';
    if(SetEvent(connector.h_thEnd))
    {
      cout<<"     Waiting for thread to terminate."<<'\n';
      WaitForSingleObject(connector.handle,INFINITE);
      CloseHandle(connector.handle);
      cout<<"     Thread terminated normally."<<'\n';
      host.stop();
      cout<<"     New connection port is closed."<<'\n';
    }
    else
    {
      // force termination of thread
      TerminateThread(connector.handle,1);
      cout<<"     Forcing termination of thread."<<'\n';
      host.stop();
      CloseHandle(connector.handle);
      CloseHandle(connector.h_thEnd);
      cout<<"     New connection port is closed.";
    }

    // Kill the SENDOR thread
    cout<<" *** Terminating SENDOR thread."<<'\n';
    if(SetEvent(sendor.h_thEnd))
    {
      cout<<"     Waiting for thread to terminate."<<'\n';
      WaitForSingleObject(sendor.handle,INFINITE);
      CloseHandle(sendor.handle);
      cout<<"     Thread terminated normally."<<'\n';
      if(modelconnected)
        cout<<"     Connection to calling program is closed."<<'\n';
    }
    else
    {
      // force termination of thread
      cout<<"     Forcing termination of thread."<<'\n';
      TerminateThread(sendor.handle,1);
      CloseHandle(sendor.handle);
      CloseHandle(sendor.h_thEnd);
      if(modelconnected)
        cout<<"     Connection to calling program is closed."<<'\n';
    }

    // Call terminator to kill slave threads
    cout<<" *** Terminating RECEIVOR threads."<<'\n';
    m_wait=WaitForSingleObject(m_thTerm,INFINITE);
    if(m_wait==WAIT_OBJECT_0)
    {
      cout<<"     Prepping queues for TERMINATOR thread."<<'\n';
      terminator.set_threads_end(*threads);
      if(!ReleaseMutex(m_thTerm)) throw ERROR_EVENT_RELEASE;
    }
    else
      cout<<" *** UNEXPECTED ERROR WAITING FOR TERMINATOR MUTEX."<<'\n';
    if(!SetEvent(h_thTerm))
    {
      // terminate threads manually
      for(thread=threads->begin();thread!=threads->end();++thread)
      {
        TerminateThread((*thread)->h_thEnd,0);
        CloseHandle((*thread)->h_thEnd);
      }
      cout<<"     Forced termination of all RECEIVOR threads."<<'\n';
    }
    else
      cout<<"     Signaled TERMINATOR thread..."<<'\n';

    // Terminate the Calling program thread
    cout<<" *** Terminating CALLING PROGRAM thread."<<'\n';
    if(SetEvent(modelthread->get_parent_end()))
    {
      cout<<"     Waiting for thread to terminate."<<'\n';
      WaitForSingleObject(modelthread->handle,INFINITE);
      CloseHandle(modelthread->handle);
      cout<<"     Thread terminated normally."<<'\n';
    }
    else
    {
      // force termination of thread
      cout<<"     Forcing termination of thread."<<'\n';
      TerminateThread(modelthread->handle,1);
      CloseHandle(modelthread->handle);
    }
    CloseHandle(modelthread->get_parent_end());

    // Terminate the TERMINATOR thread
    cout<<" *** Terminating TERMINATOR thread."<<'\n';
    if(SetEvent(terminator.h_thEnd))
    {
      cout<<"     Waiting for thread to terminate."<<'\n';
      WaitForSingleObject(terminator.handle,INFINITE);
      CloseHandle(terminator.handle);
      cout<<"     Thread terminated normally."<<'\n';
    }
    else
    {
      // force termination of thread
      cout<<"     Forcing termination of thread."<<'\n';
      TerminateThread(terminator.handle,1);
      CloseHandle(terminator.handle);
      CloseHandle(terminator.h_thEnd);
    }

    // Disconnect all slaves
    cout<<" *** Disconnecting SLAVES."<<'\n';
    rm.terminate_clients();
    clients_new->clear();
    delete clients_new;
    clients_connected->clear();
    delete clients_connected;

    delete modelclient;

    // Clear all thread queues - terminator thread deletes them
    cout<<" *** Clearing THREAD queues."<<'\n';
    threads->clear();
    delete threads;
    threads_end->clear();
    delete threads_end;

    // Runs are deleted by the RUN MANAGER, only need to clear run queues
    cout<<" *** Clearing RUN queues."<<'\n';
    runs_new->clear();
    delete runs_new;
    runs_complete->clear();
    delete runs_complete;

    // Results are deleted by SENDOR once they are sent to the calling program
    // only need to clear result queues
    results_new->clear();
    delete results_new;
    results_tosend->clear();
    delete results_tosend;

    // Close all handles
    cout<<" *** Close all EVENT handles."<<'\n';
    CloseHandle(h_thSend);
    CloseHandle(h_thTerm);
    CloseHandle(h_thComm);
    CloseHandle(m_thSlaves);
    CloseHandle(m_thRuns);
    CloseHandle(m_thSend);
    CloseHandle(m_thTerm);

    if(WSACleanup()==SOCKET_ERROR) err=ERROR_WSACLEANUP;
    switch(err)
    {
    case NORMAL_TERMINATION:
      cout<<"\n\n Normal Termination of Program."<<'\n';
      break;
    case ERROR_WSACLEANUP:
      cout<<"\n\n Winsock function 'WSACleanup' failed with error: "<<WSAGetLastError()<<'\n';
      cout<<"\n *** Termination of program with errors."<<'\n';
      break;
    case ERROR_WSASTARTUP:
      cout<<"\n\n Winsock function 'WSAStartup' failed with error: "<<WSAGetLastError()<<'\n';
      cout<<"\n *** Termination of program with errors."<<'\n';
      break;
    case ERROR_EVENT_INIT:
      cout<<"\n\n   Error initializing EVENT"<<'\n';
      cout<<"\n *** Termination of program with errors."<<'\n';
      break;
    case ERROR_EVENT_WAIT:
      cout<<"\n\n   Unexpected error waiting for EVENT"<<'\n';
      cout<<"\n *** Termination of program with errors."<<'\n';
      break;
    case ERROR_EVENT_RELEASE:
      cout<<"\n\n   Unexpected error releasing EVENT mutex"<<'\n';
      cout<<"\n *** Termination of program with errors."<<'\n';
      break;
    case ERROR_EVENT_SET:
      cout<<"\n\n   Unexpected error raising EVENT"<<'\n';
      cout<<"\n *** Termination of program with errors."<<'\n';
      break;
    case ERROR_NODE_INIT:
      cout<<"\n *** Error initializing HOST."<<'\n';
      cout<<"\n *** Termination of program with errors."<<'\n';
      break;
    case ERROR_NODE_START:
      cout<<"\n *** Error starting HOST."<<'\n';
      cout<<"\n *** Termination of program with errors."<<'\n';
      break;
     case ERROR_THREAD_START:
      cout<<"\n *** Error starting THREAD."<<'\n';
      cout<<"\n *** Termination of program with errors."<<'\n';
      break;
   }

    return 0;
  }

}