/*
 Copyright 2012, S.S. Papadopulos & Assoc. Inc.

    This file is part of GENIE.

    The GENIE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    GENIE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GENIE.  If not, see<http://www.gnu.org/licenses/>.
*/

// cmuffels, Genie version 1, 2012

#include <iostream>

#include "host.h"
#include "commandline.h"
#include "socket_utilities.h"
#include "defin.h"

using namespace std;

typedef struct addrinfo addr;

//****************************************************************************
int HOST::initialize(int &_argc,
                     char *_argv[])
//****************************************************************************
//CTM Dec 2010
/*
      Routine to initialize host properties
*/

{

  int ierr;
  COMMANDLINE cmd;
  SOCKET_UTILITIES sock_util;
  string s_sock;

  try
  {

    //pass commandline arguments to cmd
    cmd.setarg(_argc,_argv);

    //initialize socket (ip and port)
    s_sock=cmd.getarg("/sock");
    if(s_sock.compare("NOT_FOUND")==0)
    {
      //check for ip
      myIP=cmd.getarg("/ip");
      if(myIP.compare("NOT_FOUND")==0)
      {
        myIP=sock_util.ip_automatic();
        if(myIP.compare("NOT_SET")==0) throw -2;
      }
      //check for port
      myStrPORT=cmd.getarg("/port");
      if(myStrPORT=="NOT_FOUND")
      {
        myPORT=sock_util.port_automatic();
        if(myPORT==NOT_SET) throw -3;
        myStrPORT=sock_util.strport(myPORT);
      }
      else
        myPORT=sock_util.uintport(myStrPORT);
    }
    else
    {
      ierr=sock_util.parsesocket(s_sock,myIP,myPORT);
      myStrPORT=sock_util.strport(myPORT);
      if(ierr<0) throw -4;
    }
    cout<<"\n\n HOST to use socket: "<<myIP<<":"<<myPORT<<'\n';

    //initialize additional variables
    terminate=false;
    //name="HOST";
    //FD_ZERO(&fdset);
    //tolc=clock();
    //mybool=_bool;
    //mymutex=_mutex;
//    runs=_runs;
    //cout<<"RUNS SIZE: "<<runs->size()<<'\n';
    //wait=COMM_WAIT_TIME;

    return 1;
  }
  catch (int err)
  {
    if(err==-2)
    {
      cout<<" *** Please verify internet connection or,"<<'\n';
      cout<<"     Specify IP using '/ip' switch. See manual for more details."<<'\n';
    }
    else if(err==-3)
    {
      cout<<"\n\n Unable to determine PORT."<<'\n';
      cout<<" *** Please verify internet connection or,"<<'\n';
      cout<<"     Specify PORT using '/port' switch. See manual for more details."<<'\n';
    }
    else if(err==-4)
      cout<<"\n\n Error parsing supplied socket for IP and PORT."<<'\n';

    return 0;
  }

}

//****************************************************************************
int HOST::start()
//****************************************************************************
//CTM Dec 2010

/*
      Routine to open, bind and start listening on host socket
*/

{

  //routine specific variables
  int ierr;
  addr *sa,hint;
  sockaddr_in serverINF;
  int flag=1;

  //parse socket into port and ip strings
  try
  {
    //initialize hint structure
    memset(&hint,0,sizeof(hint));
    hint.ai_family=AF_INET;
    hint.ai_socktype=SOCK_STREAM;
    hint.ai_protocol=IPPROTO_TCP;
    hint.ai_flags=AI_PASSIVE;

    //initialize sa and convert PORT to address type
    if(getaddrinfo(myIP.c_str(),myStrPORT.c_str(),&hint,&sa)!=0) throw -1;

    //open socket
    sock=socket(sa->ai_family,sa->ai_socktype,sa->ai_protocol);
    if(sock==INVALID_SOCKET) throw -2;

    //make socket non-blocking and keep it alive
    if(!makenonblocking()) throw -6;
    //if(!keepalive()) throw -6;

    //set exclusive bind option
    if(setsockopt(sock,SOL_SOCKET,SO_EXCLUSIVEADDRUSE,(char *)&flag,sizeof(int))!=0) throw -3;
    if(setsockopt(sock,SOL_SOCKET,SO_KEEPALIVE,(char *)&flag,sizeof(int))!=0) throw -3;

    //bind socket to port
    serverINF.sin_family=AF_INET;
    serverINF.sin_addr.s_addr=INADDR_ANY;
    serverINF.sin_port=htons(myPORT);
    ierr=bind(sock,(SOCKADDR*)&serverINF,sizeof(serverINF));
    if(ierr==SOCKET_ERROR) throw -4;

    // set socket to listen state
    // why does SOMAXCONN not work?
    ierr=listen(sock,MAX_QUEUE_SIZE);
    if(ierr!=0) throw -5;

    cout<<" *** HOST connected to socket."<<'\n';

    return 1;

  }
  catch (int err)
  {
    if(err==-1)
      cout<<"\n\n Error converting IP and PORT to ADDRESS."<<'\n';
    if(err==-2)
      cout<<"\n\n Error opening SOCKET."<<'\n';
    if(err==-3)
      cout<<"\n\n Error setting SOCKET option."<<'\n';
    if(err==-4)
      cout<<"\n\n Error binding SOCKET."<<'\n';
    if(err==-5)
      cout<<"\n\n Error setting SOCKET to LISTEN."<<'\n';
    if(err==-6)
      cout<<" *** Error setting SOCKET option."<<'\n';

    return 0;
  }
}

//**********************************************************************
void HOST::enablesocket(SOCKET &_socket)
//**********************************************************************
//CTM, Dec 2010

/*
      Routine to add socket to master fd set
*/

{
  //FD_SET(_socket,&fdset);
}

//**********************************************************************
void HOST::disablesocket(SOCKET &_socket)
//**********************************************************************
//CTM, Dec 2010

/*
      Routine to remove socket from master fd set
*/

{
  //FD_CLR(_socket,&fdset);
}

//**********************************************************************
SOCKET HOST::acceptnewconnection()
//**********************************************************************
//CTM, Dec 2010

/*
      Routine to accept new connection
*/

{
  
  SOCKET newsock=SOCKET_ERROR;

  //try
  //{

    //accept socket
    while(newsock==SOCKET_ERROR)        //i may want to take this out of a loop and just attempt it once
      newsock=accept(sock,NULL,NULL);
    return newsock;

  //}
  //catch (...)
  //{
  //  cout<<"\n\n Error accepting connection. Ignoring."<<'\n';
  //  return newsock;
  //}

}

//****************************************************************************
int HOST::waitforcommunication(long sec,long usec)
//****************************************************************************
//CTM Dec 2010
// from http://www.winsocketdotnetworkprogramming.com/winsock2programming/
//                                             winsock2advancedcode1c.html

/*
      Routine to wait sec and usec long for activity on a socket
*/
{

  // Setup timeval variable
  struct timeval timeout;
  fd_set _fdset;

  // assign socket to fdset
  FD_ZERO(&_fdset);
  FD_SET(sock,&_fdset);

   // assign the second and microsecond variables
  timeout.tv_sec = sec;
  timeout.tv_usec = usec;

  // Possible return values:
  // -1: error occurred
  // 0: timed out
  // > 0: data ready to be read
  if(sec==-1)
    return select(FD_SETSIZE, &_fdset, NULL, NULL, NULL);
  else
    return select(FD_SETSIZE, &_fdset, NULL, NULL, &timeout);

}