/*
 Copyright 2012, S.S. Papadopulos & Assoc. Inc.

    This file is part of GENIE.

    The GENIE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    GENIE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GENIE.  If not, see<http://www.gnu.org/licenses/>.
*/

// cmuffels, Genie version 1, 2012

#include <iostream>

#include "managedrun.h"

using namespace std;

//****************************************************************************
void MANAGED_RUN::set_run(MODEL_RUN *_run)
//****************************************************************************
//CTM Apr 2011

/*
      Set model run associated with this managed run
*/

{
  run=_run;
}

//****************************************************************************
void MANAGED_RUN::set_status(int _status)
//****************************************************************************
//CTM Apr 2011

/*
      Set status of run
*/

{
  status=_status;
}

//****************************************************************************
size_t MANAGED_RUN::client_count()
//****************************************************************************
//CTM Apr 2011

/*
      Return the number of slaves executing this run
*/

{

  return slaves.size();

}

//****************************************************************************
size_t MANAGED_RUN::id()
//****************************************************************************
//CTM Apr 2011

/*
      Return run id
*/

{

  return *run->id;

}

//****************************************************************************
int MANAGED_RUN::add_slave(CLIENT *_slave)
//****************************************************************************
//CTM Apr 2011

/*
      Add slave to list of slaves executing this run
*/

{

  try
  {
    _slave->run_start->push_back(clock());
    if(!run->send(_slave->sock)) throw ERROR_RUN_SEND;
    slaves.insert(_slave);
    return 1;
  }
  catch(...)
  {
    cout<<"\n\n   Error sending RUN: "<<run->id<<" to CLIENT: "<<_slave->name<<'\n';
    return 0;
  }
}

//****************************************************************************
void MANAGED_RUN::check_client_status()
//****************************************************************************
//CTM Apr 2011

/*
      Process slave according to its status
*/

{

  set<CLIENT*>::iterator slave;

  slave=slaves.begin();
  while(slave!=slaves.end())
  {
    if((*slave)->status==CLIENT_LOST)
      slaves.erase(slave++);
    else
    {
      if((*slave)->status==RUN_COMPLETE)
        status=MRUN_COMPLETE;
      ++slave;
    }
  }

}

//****************************************************************************
void MANAGED_RUN::remove_slave(CLIENT *_slave)
//****************************************************************************
//CTM Apr 2011

/*
      Remove slave from list of slaves executing this run
*/

{

  slaves.erase(_slave);

}

//****************************************************************************
void MANAGED_RUN::kill_runs(multiset<CLIENT*,cmp_speed> *_clients_waiting,
                            multiset<CLIENT*,cmp_speed> *_clients_working)
//****************************************************************************
//CTM Apr 2011

/*
      Send kill command to slaves associated with this run
*/

{

  set<CLIENT*>::iterator slave;
  multiset<CLIENT*,cmp_speed>::iterator islv;

  for(slave=slaves.begin();slave!=slaves.end();++slave)
  {
    if((*slave)->status!=RUN_COMPLETE)
    {
      (*slave)->run_start->pop_back();
      if(!(*slave)->sendcommand(KILL_RUN))
      {
        cout<<"I think I lost slave: "<<(*slave)->name<<'\n';
        (*slave)->status=CLIENT_LOST;
      }
    }
    // update slave status
    (*slave)->status=CLIENT_WAITING;
    // update queues
    _clients_waiting->insert(*slave);
    islv=_clients_working->find(*slave);
    if(islv==_clients_working->end())
      cout<<"Error finding "<<(*slave)->name<<" in list of working slaves."<<'\n';
    else
      _clients_working->erase(islv);
  }
  slaves.clear();

}
