/*
 Copyright 2012, S.S. Papadopulos & Assoc. Inc.

    This file is part of GENIE.

    The GENIE is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    GENIE is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with GENIE.  If not, see<http://www.gnu.org/licenses/>.
*/

// cmuffels, Genie version 1, 2012

#include <iostream>

#include "terminator.h"

using namespace std;

//****************************************************************************
void TERMINATOR::set_threads_end(list<RECEIVOR*> _threads_end)
//****************************************************************************
//CTM Apr 2011

/*
      Set list of threads to terminate
*/

{
  threads_end=_threads_end;
}

//****************************************************************************
void TERMINATOR::set_handles_end(list<HANDLE> _handles_end)
//****************************************************************************
//CTM Apr 2011

/*
      Set list of end handles for threads to terminate
*/

{
  handles_end=_handles_end;
}
//
////****************************************************************************
//void TERMINATOR::set_primary_thread_handle(HANDLE _primary)
////****************************************************************************
////CTM Apr 2011
//
///*
//      Set the handle of the primary thread for use by duplicatehandle
//*/
//
//{
//  primary=_primary;
//}

//****************************************************************************
int TERMINATOR::execute()
//****************************************************************************
//CTM Apr 2011

/*
      Execute thread process
*/

{

  DWORD tr_Wait,m_Wait;
  bool endthread=false;
  HANDLE trEvents[2];
  HANDLE h_E_Dup;
  HANDLE h_T_Dup;

  list<RECEIVOR*>::iterator ith;
  list<HANDLE>::iterator ih;

  try
  {

    // duplicate handles of primary thread
    if(!DuplicateHandle(primary,ph_thComm,GetCurrentProcess(),&h_thComm,
                        0,true,DUPLICATE_SAME_ACCESS)) throw -1;
    if(!DuplicateHandle(primary,pm_thComm,GetCurrentProcess(),&m_thComm,
                        0,true,DUPLICATE_SAME_ACCESS)) throw -2;
    if(!DuplicateHandle(primary,ph_thEnd,GetCurrentProcess(),&h_thEnd,
                        0,true,DUPLICATE_SAME_ACCESS)) throw -3;
    trEvents[0]=h_thComm;
    trEvents[1]=h_thEnd;

    while(1)
    {

      if(endthread)
      {
        CloseHandle(h_thEnd);
        break;
      }

      // wait for either the execute or terminate flags to be raised
      //cout<<"waiting for threads to terminate."<<'\n';
      tr_Wait=WaitForMultipleObjects(2,trEvents,FALSE,INFINITE);
      switch(tr_Wait)
      {
      case WAIT_OBJECT_0+0:
        // Process list of threads to terminate
        //cout<<"waiting for term mutex"<<'\n';
        m_Wait=WaitForSingleObject(m_thComm,INFINITE);
        switch(m_Wait)
        {
        case WAIT_OBJECT_0:
          // trigger end for each thread
          //cout<<"terminating "<<threads_end.size()<<" threads."<<'\n';
          //ih=handles_end.begin();
          ith=threads_end.begin();
          while(ith!=threads_end.end())
          {
            //cout<<"raising terminate flag for this thread."<<'\n';
            if(!DuplicateHandle(primary,(*ith)->handle,GetCurrentProcess(),
                                                                 &h_T_Dup,0,false,DUPLICATE_SAME_ACCESS))
              cout<<"Error duplicating handle to thread to be terminated: "<<GetLastError()<<'\n';
            if(!DuplicateHandle(primary,(*ith)->get_parent_end(),GetCurrentProcess(),&h_E_Dup,0,true,DUPLICATE_SAME_ACCESS))
              cout<<"Error duplicating handle to terminate event: "<<GetLastError()<<'\n';
            //h_E_Dup=OpenEvent(0,false,(*ith)->slave->name.c_str());
            if(SetEvent(h_E_Dup))
            {
              //cout<<"waiting for thread to terminate."<<'\n';
              tr_Wait=WaitForSingleObject(h_T_Dup,COMM_WAIT_TIME*CLOCKS_PER_SEC*10);
              switch(tr_Wait)
              {
              case WAIT_OBJECT_0:
                //cout<<"normal termination of thread"<<'\n';
                CloseHandle(h_T_Dup);
                delete *ith;
                ith=threads_end.erase(ith);
                break;
              default:
                TerminateThread(h_T_Dup,0);
                //cout<<"forced termination of thread."<<'\n';
                CloseHandle(h_T_Dup);
                delete *ith;
                ith=threads_end.erase(ith);
                break;
              }
            }
            else
            {
              cout<<"Error raising END flag for Slave."<<'\n';
              cout<<"Error: "<<GetLastError()<<'\n';
              TerminateThread(h_T_Dup,0);
              CloseHandle(h_T_Dup);
              delete *ith;
              ith=threads_end.erase(ith);
              cout<<"forced termination of thread."<<'\n';
            }
          }
          if(!ReleaseMutex(m_thComm)) throw ERROR_EVENT_RELEASE;
          break;
        default:
          throw ERROR_EVENT_WAIT;
          break;
        }
        threads_end.clear();
        handles_end.clear();
        break;
      case WAIT_OBJECT_0+1:
        // Terminate thread
        if(!ResetEvent(h_thEnd)) throw ERROR_EVENT_RELEASE;
        endthread=true;
        break;
      default:
        throw ERROR_EVENT_WAIT;
        break;
      }

    }
    return NORMAL_TERMINATION;
  }

  catch(int err)
  {
    switch(err)
    {
    case -1:
      cout<<"\n\n   Error duplicating misc. communication flag."<<'\n';
      break;
    case -2:
      cout<<"\n\n   Error duplicating misc. shared access flag."<<'\n';
      break;
    case -3:
      cout<<"\n\n   Error duplicating termination flag."<<'\n';
      break;
    }
    cout<<" *** THREAD :"<<threadid<<'\n';
    return 0;
  }

}


